package ru.dbotthepony.mc.otm.network

import it.unimi.dsi.fastutil.ints.Int2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.ints.Int2ObjectArrayMap
import it.unimi.dsi.fastutil.ints.IntAVLTreeSet
import it.unimi.dsi.fastutil.ints.IntArrayList
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.inventory.Slot
import net.minecraft.world.item.ItemStack
import net.minecraftforge.network.NetworkDirection
import net.minecraftforge.network.NetworkEvent
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.block.entity.storage.ItemMonitorPlayerSettings
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.container.ItemFilterSlotPacket
import ru.dbotthepony.mc.otm.core.TextComponent
import ru.dbotthepony.mc.otm.menu.*
import ru.dbotthepony.mc.otm.menu.data.*
import ru.dbotthepony.mc.otm.menu.widget.BooleanPlayerInputPacket
import ru.dbotthepony.mc.otm.menu.widget.NumberPlayerInputPacket
import ru.dbotthepony.mc.otm.menu.widget.OneWayPlayerInputPacket
import java.io.ByteArrayInputStream
import java.util.function.Supplier

class MenuFieldPacket(val bytes: ByteArray, val length: Int) : MatteryPacket {
	constructor(stream: FastByteArrayOutputStream) : this(stream.array, stream.length)

	override fun write(buff: FriendlyByteBuf) {
		buff.writeBytes(bytes, 0, length)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true
		context.get().enqueueWork {
			(minecraft.player?.containerMenu as? MatteryMenu)?.mSynchronizer?.applyNetworkPayload(ByteArrayInputStream(bytes, 0, length))
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): MenuFieldPacket {
			val readable = buff.readableBytes()
			return MenuFieldPacket(ByteArray(readable).also(buff::readBytes), readable)
		}
	}
}

class SetCarriedPacket(val item: ItemStack) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeItem(item)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true
		context.get().enqueueWork { minecraft.player?.containerMenu?.carried = item }
	}

	companion object {
		fun read(buff: FriendlyByteBuf): SetCarriedPacket {
			return SetCarriedPacket(buff.readItem())
		}
	}
}

class MoveMultipleItemsPacket(val actions: Collection<Move>, val changedSlots: Map<Int, ItemStack>, val error: Throwable? = null) : MatteryPacket {
	data class Move(val sourceSlots: Collection<Int>, val destinationSlot: Int, val count: Int) {
		fun write(buff: FriendlyByteBuf) {
			buff.writeInt(sourceSlots.size)

			for (slot in sourceSlots) {
				buff.writeInt(slot)
			}

			buff.writeInt(destinationSlot)
			buff.writeInt(count)
		}

		companion object {
			fun read(buff: FriendlyByteBuf): Move {
				val listSize = buff.readInt()
				require(listSize <= 100) { "Too many source slots! ($listSize received)" }

				val seen = IntAVLTreeSet()
				val list = IntArrayList()

				for (i in 0 until listSize) {
					val index = buff.readInt()

					if (seen.add(index)) {
						list.add(index)
					}
				}

				val destinationSlot = buff.readInt()
				val count = buff.readInt()

				return Move(list, destinationSlot, count)
			}
		}
	}

	override fun write(buff: FriendlyByteBuf) {
		buff.writeInt(actions.size)

		for (action in actions) {
			action.write(buff)
		}

		buff.writeInt(changedSlots.size)

		for ((k, v) in changedSlots) {
			buff.writeInt(k)
			buff.writeItem(v)
		}
	}

	fun execute(ply: Player, container: AbstractContainerMenu): Map<Int, ItemStack> {
		val changedSlots = Int2ObjectArrayMap<ItemStack>()

		for ((sourceSlots, destinationSlotIndex, count) in actions) {
			val destinationSlot = container.slots[destinationSlotIndex]

			if (!destinationSlot.item.isEmpty) {
				// clear slot first
				continue
			}

			val realSourceSlots = sourceSlots.map { container.slots[it] }

			if (realSourceSlots.isEmpty() || realSourceSlots.all { it.item.isEmpty }) {
				continue
			}

			val type = realSourceSlots.first { !it.item.isEmpty }.item

			if (!realSourceSlots.all { it.item.isEmpty || it.item.item === type.item }) {
				continue
			}

			if (!destinationSlot.mayPlace(type)) {
				continue
			}

			var pickedUp = 0
			val maxPickup = destinationSlot.getMaxStackSize(type).coerceAtMost(count)

			for (slot in realSourceSlots) {
				if (slot.mayPickup(ply) && !slot.item.isEmpty) {
					if (!destinationSlot.item.isEmpty && !ItemStack.isSameItemSameTags(destinationSlot.item, slot.item)) {
						continue
					}

					val old = pickedUp
					pickedUp = (pickedUp + slot.item.count).coerceAtMost(maxPickup)
					val diff = pickedUp - old

					val copy = slot.item.copy().also { it.count = diff }

					slot.item.count -= diff

					if (slot.item.isEmpty) {
						slot.set(ItemStack.EMPTY)
						changedSlots[slot.index] = ItemStack.EMPTY
					} else {
						slot.setChanged()
						changedSlots[slot.index] = slot.item.copy()
					}

					if (destinationSlot.item.isEmpty) {
						destinationSlot.set(copy)
					} else {
						destinationSlot.item.count += diff
						destinationSlot.setChanged()
					}

					if (pickedUp == maxPickup) {
						break
					}
				}
			}

			if (pickedUp != 0) {
				changedSlots[destinationSlot.index] = destinationSlot.item.copy()
			}
		}

		return changedSlots
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true

		if (error != null) {
			context.sender!!.connection.disconnect(TextComponent("Caught an exception processing ${MoveMultipleItemsPacket::class.qualifiedName}: $error"))
			LOGGER.error("Processing multi move input from player ${context.sender} has thrown an exception!", error)
			return
		}

		context.enqueueWork {
			try {
				val container = context.sender!!.containerMenu

				if (container is ExoSuitInventoryMenu) {
					execute(context.sender!!, container)

					for ((index, contents) in changedSlots) {
						container.setRemoteSlot(index, contents)
					}
				}
			} catch (err: Throwable) {
				context.sender!!.connection.disconnect(TextComponent("Caught an exception processing ${MoveMultipleItemsPacket::class.qualifiedName}: $err"))
				LOGGER.error("Processing multi move input from player ${context.sender} has thrown an exception!", err)
			}
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()

		fun read(buff: FriendlyByteBuf): MoveMultipleItemsPacket {
			try {
				val actionCount = buff.readInt()
				require(actionCount <= 40) { "Too many actions!" }

				val actions = ArrayList<Move>(actionCount)

				for (i in 0 until actionCount) {
					actions.add(Move.read(buff))
				}

				val changedSlotCount = buff.readInt()
				require(changedSlotCount <= 40) { "Too many changed slots!" }

				val changedSlots = Int2ObjectAVLTreeMap<ItemStack>()

				for (i in 0 until changedSlotCount) {
					changedSlots.put(buff.readInt(), buff.readItem())
				}

				return MoveMultipleItemsPacket(actions, changedSlots)
			} catch (err: Throwable) {
				return MoveMultipleItemsPacket(listOf(), mapOf(), err)
			}
		}
	}
}

object MenuNetworkChannel : MatteryNetworkChannel(
	version = "1",
	name = "menu"
) {
	fun register() {
		add(SetCarriedPacket::class.java, SetCarriedPacket.Companion::read, NetworkDirection.PLAY_TO_CLIENT)
		add(ItemFilterSlotPacket::class.java, ItemFilterSlotPacket.Companion::read)

		// networked view
		add(ClearItemViewPacket::class.java, ClearItemViewPacket::read, NetworkDirection.PLAY_TO_CLIENT)
		add(ItemViewInteractPacket::class.java, ItemViewInteractPacket.Companion::read, NetworkDirection.PLAY_TO_SERVER)
		add(StackAddPacket::class.java, StackAddPacket.Companion::read, NetworkDirection.PLAY_TO_CLIENT)
		add(StackChangePacket::class.java, StackChangePacket.Companion::read, NetworkDirection.PLAY_TO_CLIENT)
		add(StackRemovePacket::class.java, StackRemovePacket.Companion::read, NetworkDirection.PLAY_TO_CLIENT)

		// container data
		add(MenuFieldPacket::class.java, MenuFieldPacket.Companion::read, NetworkDirection.PLAY_TO_CLIENT)

		// Player input in menus
		add(NumberPlayerInputPacket::class.java, NumberPlayerInputPacket.Companion::read, NetworkDirection.PLAY_TO_SERVER)
		add(OneWayPlayerInputPacket::class.java, OneWayPlayerInputPacket.Companion::read, NetworkDirection.PLAY_TO_SERVER)
		add(BooleanPlayerInputPacket::class.java, BooleanPlayerInputPacket.Companion::read, NetworkDirection.PLAY_TO_SERVER)

		// menu specific
		add(MoveMultipleItemsPacket::class.java, MoveMultipleItemsPacket.Companion::read, NetworkDirection.PLAY_TO_SERVER)

		// Item monitor
		add(ItemMonitorPlayerSettings::class.java, ItemMonitorPlayerSettings.Companion::read)

		// matter panel menu
		add(CancelTaskPacket::class.java, CancelTaskPacket.Companion::read, NetworkDirection.PLAY_TO_SERVER)
		add(PatternsChangePacket::class.java, PatternsChangePacket.Companion::read, NetworkDirection.PLAY_TO_CLIENT)
		add(TasksChangePacket::class.java, TasksChangePacket.Companion::read, NetworkDirection.PLAY_TO_CLIENT)
		add(ReplicationRequestPacket::class.java, ReplicationRequestPacket.Companion::read, NetworkDirection.PLAY_TO_SERVER)
	}
}
